
/***************************************************************************
 *
 * ArrayBinaryHeap.java
 *
 * The ArrayBinaryHeap.java class implements a binary heap.
 * Note that all "matching" is based on the compareTo method.
 *
 *****************************************************************************/

import java.util.*;


public class ArrayBinaryHeap<T extends Comparable<? super T>> implements BinaryHeap<T>
{

    private static final int DEFAULT_CAPACITY = 10;

    private int currentSize;      // Number of elements in heap
    private T [ ] array; 		  // The heap array


    /**
     * Construct the binary heap.
     */
    public ArrayBinaryHeap( )
    {
        this( DEFAULT_CAPACITY );
    }


    /**
     * Construct the binary heap.
     * @param the capacity of the binary heap.
     */
    public ArrayBinaryHeap( int capacity )
    {
    	currentSize = 0;
        array = (T[]) new Comparable<?>[ capacity + 1 ];
    }
    
    
	
	/**
     * Insert into the array, maintaining heap order.
     * Duplicates are allowed.
     * @param x the item to insert.
     */
    public void insert( T x )
    {
        if( currentSize == array.length - 1 )
            reSizeArray( array.length * 2 );

        // Fix up
        int hole = ++currentSize;
        for( array[ 0 ] = x; x.compareTo( array[ hole / 2 ] ) < 0; hole /= 2 )
            array[ hole ] = array[ hole / 2 ];
        array[ hole ] = x;
    }


    private void reSizeArray( int newSize )
    {
            T [] old = array;
            if (newSize <= currentSize + 1) return;
            array = (T []) new Comparable<?>[ newSize ];
            for( int i = 0; i < currentSize+1; i++ )
                array[ i ] = old[ i ];        
    }
    
   	
	/**
     * Internal method to percolate down in the heap.
     * @param hole the index at which the percolate begins.
     */
    private void fixHeap( int hole )  // Fix down
    {
        int child;
        T tmp = array[ hole ];

        for( ; hole * 2 <= currentSize; hole = child )
        {
            child = hole * 2;
            if( child != currentSize &&
                    array[ child + 1 ].compareTo( array[ child ] ) < 0 )
                child++;
            if( array[ child ].compareTo( tmp ) < 0 )
                array[ hole ] = array[ child ];
            else
                break;
        }
        array[ hole ] = tmp;
    }

    
    /**
     * Establish heap order property from an arbitrary
     * arrangement of items. Runs in linear time.
     */
    public void heapify( )
    {
        for( int i = currentSize / 2; i > 0; i-- )
            fixHeap( i );  // Fix down
    }

	
	
   /**
     * Construct the binary heap given an array of items.
     */
    public ArrayBinaryHeap( T [ ] items )
    {
            currentSize = items.length;
            array = (T[]) new Comparable<?>[ currentSize + 1 ];
            int i = 1;
            for( T item : items ) array[ i++ ] = item;
            heapify( );
    }


    
    /**
     * Find the smallest item in the heap.
     * @return the smallest item, or throw an exception if empty.
     */
    public T findMin( )
    {
        if( isEmpty( ) )
		throw new NoSuchElementException("The binary heap is empty");
        return array[1];
    }

	
    /**
     * Remove the smallest item from the priority queue.
     * @return the smallest item, or throw an exception if empty.
     */
    public T deleteMin( )
    {
        if( isEmpty( ) )
		throw new NoSuchElementException("The binary heap is empty");
        T minItem = array[1];  //findMin( );
        array[1] = array[ currentSize-- ];
        fixHeap( 1 );  // Fix down
        if (currentSize +1 <= array.length/4)
        	reSizeArray( array.length / 2);
        return minItem;
    }


    /**
     * Test if the heap is logically empty.
     * @return true if empty, false otherwise.
     */
    public boolean isEmpty( )
    {
        return currentSize == 0;
    }

    /**
     * Make the heap logically empty.
     */
    public void clear( )
    {
        currentSize = 0;
    }
    
    public int size() {return currentSize;}


}
